use serde::Deserialize;
use tracing::debug;

#[derive(Debug, Clone)]
pub(crate) struct GiteaClient {
    api_root: String,
    token: String,
    http_client: reqwest::Client,
}

impl GiteaClient {
    pub(crate) fn new<R: Into<String>, T: Into<String>>(api_root: R, token: T) -> GiteaClient {
        GiteaClient {
            api_root: api_root.into(),
            token: token.into(),
            http_client: reqwest::Client::new(),
        }
    }

    pub(crate) async fn get_tree<O: AsRef<str>, R: AsRef<str>, C: AsRef<str>>(
        &self,
        owner: O,
        repo: R,
        commit: C,
    ) -> Result<Tree, Box<dyn std::error::Error>> {
        let mut files = Vec::new();
        let mut page: Option<u64> = None;
        loop {
            let url = format!(
                "{api_root}/v1/repos/{owner}/{repo}/git/trees/{commit}?recursive=true&per_page=10{page}",
                api_root = self.api_root,
                owner = owner.as_ref(),
                repo = repo.as_ref(),
                commit = commit.as_ref(),
                page = page.map(|num| format!("&page={}", num)).unwrap_or_else(|| String::new())
            );
            let response = self
                .http_client
                .get(url)
                .header("Authorization", format!("token {}", self.token))
                .send()
                .await?;
            let response = response.error_for_status()?;
            let body = response.text().await?;
            debug!("Response: {}", body);
            let parsed_body: ResponseGetTree = serde_json::from_str(body.as_str())?;
            println!("Response: {:#?}", parsed_body);
            // TODO: Read response objects

            if !parsed_body.truncated {
                break;
            }
            // TODO: Switch to using response headers instead of response body.
            page = Some(parsed_body.page + 1);
        }
        Ok(Tree::new(files))
    }
}

/// A single API response for GetTree containing only one page.
#[derive(Debug, Deserialize)]
struct ResponseGetTree {
    sha: String,
    url: String,
    tree: Vec<ResponseObjectReference>,
    truncated: bool,
    page: u64,
    total_count: u64,
}

#[derive(Debug, Deserialize)]
struct ResponseObjectReference {
    path: String,
    mode: String,
    #[serde(rename = "type")]
    object_type: String,
    size: u64,
    sha: String,
    url: String,
}

#[derive(Debug)]
pub(crate) struct Tree {
    pub(crate) files: Vec<TreeFileReference>,
}

#[derive(Debug)]
pub(crate) struct TreeFileReference {}

impl Tree {
    pub(crate) fn new(files: Vec<TreeFileReference>) -> Tree {
        Tree { files: Vec::new() }
    }
}

impl TreeFileReference {
    pub(crate) fn new() -> TreeFileReference {
        TreeFileReference {}
    }
}
